import re
import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional
from importlib.resources import files as resource_file

from lxml.etree import (
    Element,
    ElementTree,
    tostring as element_string,
    SubElement,
)  # nosec - see 'lxml` package (bandit)' section in README


fpl_waypoint_types = ["USER WAYPOINT", "AIRPORT", "NDB", "VOR", "INT", "INT-VRP"]


def _upper_alphanumeric_space_only(value: str) -> str:
    return re.sub(r"[^A-Z\d ]+", "", value.upper())


def _upper_alphanumeric_only(value: str) -> str:
    return re.sub(r"[^A-Z\d]+", "", value.upper())


class Namespaces(object):
    fpl = "http://www8.garmin.com/xmlschemas/FlightPlan/v1"
    xsi = "http://www.w3.org/2001/XMLSchema-instance"

    _root_namespace = fpl

    _schema_locations = {
        "fpl": "http://www8.garmin.com/xmlschemas/FlightPlanv1.xsd",
    }

    namespaces = {
        "fpl": fpl,
        "xsi": xsi,
    }

    @staticmethod
    def nsmap(suppress_root_namespace: bool = False) -> dict:
        """
        Create a namespace map.

        Indexes namespaces by their prefix.

        E.g. {'xlink': 'http://www.w3.org/1999/xlink'}

        When a root namespace is set, a default namespace will be set by using the `None` constant for the relevant
        dict key (this is an lxml convention). This will create an invalid namespace map for use in XPath queries, this
        can be overcome using the `suppress_root_namespace` parameter, which will create a 'regular' map.

        :type suppress_root_namespace: bool
        :param suppress_root_namespace: When true, respects a root prefix as a default if set
        :return: dictionary of Namespaces indexed by prefix
        """
        nsmap = {}

        for prefix, namespace in Namespaces.namespaces.items():
            if namespace == Namespaces._root_namespace and not suppress_root_namespace:
                nsmap[None] = namespace
                continue

            nsmap[prefix] = namespace

        return nsmap

    @staticmethod
    def schema_locations() -> str:
        """
        Generates the value for a `xsi:schemaLocation` attribute

        Defines the XML Schema Document (XSD) for each namespace in an XML tree

        E.g. 'xsi:schemaLocation="http://www.w3.org/1999/xlink https://www.w3.org/1999/xlink.xsd"'

        :rtype str
        :return: schema location attribute value
        """
        schema_locations = ""
        for prefix, location in Namespaces._schema_locations.items():
            schema_locations = f"{schema_locations} {Namespaces.namespaces[prefix]} {location}"

        return schema_locations.lstrip()


class Waypoint:
    def __init__(
        self,
        identifier: Optional[str] = None,
        waypoint_type: Optional[str] = None,
        country_code: Optional[str] = None,
        latitude: Optional[float] = None,
        longitude: Optional[float] = None,
        comment: Optional[str] = None,
    ):
        self.ns = Namespaces()

        self._identifier: Optional[str] = None
        self._type: Optional[str] = None
        self._country_code: Optional[str] = None
        self._latitude: Optional[float] = None
        self._longitude: Optional[float] = None
        self._comment: Optional[str] = None

        if identifier is not None:
            self.identifier = identifier
        if waypoint_type is not None:
            self.type = waypoint_type
        if country_code is not None:
            self.country_code = country_code
        if latitude is not None:
            self.latitude = latitude
        if longitude is not None:
            self.longitude = longitude
        if comment is not None:
            self.comment = comment

    @property
    def identifier(self) -> str:
        return self._identifier

    @identifier.setter
    def identifier(self, identifier: str) -> None:
        if identifier is not None:
            if len(identifier) > 12:
                raise ValueError("Identifier must be 12 characters or less.")
            self._identifier = _upper_alphanumeric_only(value=identifier)

    @property
    def type(self) -> str:
        return self._type

    @type.setter
    def type(self, waypoint_type: str) -> None:
        if waypoint_type not in fpl_waypoint_types:
            raise ValueError(f"Waypoint type must be one of '{' '.join(fpl_waypoint_types)}'")

        self._type = waypoint_type

    @property
    def country_code(self) -> str:
        return self._country_code

    @country_code.setter
    def country_code(self, country_code: Optional[str]) -> None:
        if country_code is not None:
            if len(country_code) > 2:
                raise ValueError("Country code must be 2 characters or less.")

            self._country_code = _upper_alphanumeric_only(value=country_code)

            # As a exception for Antarctica, we use '__' as the country code
            if country_code == "__":
                self._country_code = "__"

    @property
    def latitude(self) -> float:
        return self._latitude

    @latitude.setter
    def latitude(self, latitude: float) -> None:
        if latitude == 90:
            latitude = 89.999999
        elif latitude == -90:
            latitude = -89.999999

        self._latitude = latitude

    @property
    def longitude(self) -> float:
        return self._longitude

    @longitude.setter
    def longitude(self, longitude: float) -> None:
        if longitude == 90:
            longitude = 89.999999
        elif longitude == -90:
            longitude = -89.999999

        self._longitude = longitude

    @property
    def comment(self) -> str:
        if self._comment is None:
            return "NO COMMENT"
        return self._comment

    @comment.setter
    def comment(self, comment: str) -> None:
        if len(comment) > 25:
            raise ValueError("Comments must be 25 characters or less.")

        self._comment = _upper_alphanumeric_space_only(value=comment)

    def encode(self) -> Element:
        waypoint = Element(f"{{{self.ns.fpl}}}waypoint", nsmap=self.ns.nsmap())

        identifier = SubElement(waypoint, f"{{{self.ns.fpl}}}identifier")
        identifier.text = self.identifier

        waypoint_type = SubElement(waypoint, f"{{{self.ns.fpl}}}type")
        waypoint_type.text = self.type

        country_code = SubElement(waypoint, f"{{{self.ns.fpl}}}country-code")
        country_code.text = self.country_code

        latitude = SubElement(waypoint, f"{{{self.ns.fpl}}}lat")
        latitude.text = str(round(self.latitude, ndigits=7))

        longitude = SubElement(waypoint, f"{{{self.ns.fpl}}}lon")
        longitude.text = str(round(self.longitude, ndigits=7))

        if self.comment is not None:
            comment = SubElement(waypoint, f"{{{self.ns.fpl}}}comment")
            comment.text = self.comment

        return waypoint


class RoutePoint:
    def __init__(
        self,
        waypoint_identifier: Optional[str] = None,
        waypoint_type: Optional[str] = None,
        waypoint_country_code: Optional[str] = None,
    ):
        self.ns = Namespaces()

        self._waypoint_identifier: Optional[str] = None
        self._waypoint_type: Optional[str] = None
        self._waypoint_country_code: Optional[str] = None

        if waypoint_identifier is not None:
            self.waypoint_identifier = waypoint_identifier

        if waypoint_type is not None:
            self.waypoint_type = waypoint_type

        if waypoint_country_code is not None:
            self.waypoint_country_code = waypoint_country_code

    @property
    def waypoint_identifier(self) -> str:
        return self._waypoint_identifier

    @waypoint_identifier.setter
    def waypoint_identifier(self, waypoint_identifier: str) -> None:
        if len(waypoint_identifier) > 12:
            raise ValueError("Waypoint identifier must be 12 characters or less.")

        self._waypoint_identifier = _upper_alphanumeric_only(value=waypoint_identifier)

    @property
    def waypoint_type(self) -> str:
        return self._waypoint_type

    @waypoint_type.setter
    def waypoint_type(self, waypoint_type: str) -> None:
        if waypoint_type not in fpl_waypoint_types:
            raise ValueError(f"Waypoint type must be one of '{' '.join(fpl_waypoint_types)}'")

        self._waypoint_type = waypoint_type

    @property
    def waypoint_country_code(self) -> str:
        return self._waypoint_country_code

    @waypoint_country_code.setter
    def waypoint_country_code(self, waypoint_country_code: Optional[str]) -> None:
        if waypoint_country_code is not None:
            if len(waypoint_country_code) > 2:
                raise ValueError("Country code must be 2 characters or less.")

            self._waypoint_country_code = _upper_alphanumeric_only(value=waypoint_country_code)

            # As a exception for Antarctica, we use '__' as the country code
            if waypoint_country_code == "__":
                self._waypoint_country_code = "__"

    def encode(self) -> Element:
        route_point = Element(f"{{{self.ns.fpl}}}route-point", nsmap=self.ns.nsmap())

        waypoint_identifier = SubElement(route_point, f"{{{self.ns.fpl}}}waypoint-identifier")
        waypoint_identifier.text = self.waypoint_identifier

        waypoint_type = SubElement(route_point, f"{{{self.ns.fpl}}}waypoint-type")
        waypoint_type.text = self.waypoint_type

        waypoint_country_code = SubElement(route_point, f"{{{self.ns.fpl}}}waypoint-country-code")
        waypoint_country_code.text = self.waypoint_country_code

        return route_point


class Route:
    def __init__(self, name: Optional[str] = None, index: Optional[int] = None, points: Optional[List[dict]] = None):
        self.ns = Namespaces()

        self._name: Optional[str] = None
        self._index: Optional[int] = None
        self._points: Optional[List[RoutePoint]] = []

        if name is not None:
            self.name = name

        if index is not None:
            self.index = index

        if points is not None:
            self.points = points

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        if len(name) > 25:
            raise ValueError("Name must be 25 characters or less.")

        # FPL uses space (' ') as a separator character not underscore ('_') so replace any first.
        # TODO: This should move to the revised NetworkManager class as it's BAS specific, see #46.
        name = name.replace("_", " ")

        self._name = _upper_alphanumeric_space_only(value=name)

    @property
    def index(self) -> int:
        return self._index

    @index.setter
    def index(self, index: int) -> None:
        if index > 99:
            raise ValueError("Index must be 98 or less.")

        self._index = index

    @property
    def points(self) -> List[RoutePoint]:
        return self._points

    @points.setter
    def points(self, points: List[RoutePoint]):
        self._points = points

    def encode(self) -> Element:
        route = Element(f"{{{self.ns.fpl}}}route", nsmap=self.ns.nsmap())

        route_name = SubElement(route, f"{{{self.ns.fpl}}}route-name")
        route_name.text = self.name

        route_index = SubElement(route, f"{{{self.ns.fpl}}}flight-plan-index")
        route_index.text = str(self.index)

        if len(self.points) > 3000:
            raise ValueError("FPL routes must have 3000 points or less.")

        for route_point in self.points:
            route.append(route_point.encode())

        return route


class Fpl:
    schema_path = resource_file("bas_air_unit_network_dataset.schemas.garmin").joinpath("FlightPlanv1.xsd")

    def __init__(self, waypoints: Optional[List[Waypoint]] = None, route: Optional[Route] = None):
        self.ns = Namespaces()

        self._waypoints: List[Waypoint] = []
        self._route: Optional[Route] = None

        if waypoints is not None:
            self.waypoints = waypoints

        if route is not None:
            self.route = route

    @property
    def waypoints(self) -> List[Waypoint]:
        return self._waypoints

    @waypoints.setter
    def waypoints(self, waypoints: List[Waypoint]):
        self._waypoints = waypoints

    @property
    def route(self) -> Route:
        return self._route

    @route.setter
    def route(self, route: Route) -> None:
        self._route = route

    def dumps_xml(self) -> bytes:
        """
        Generates an XML document and tree from a root XML element

        The XML document is encoded as a UTF-8 byte string, with pretty-printing and an XML declaration.

        :rtype bytes
        :return: XML document in bytes
        """
        root = Element(
            f"{{{self.ns.fpl}}}flight-plan",
            attrib={
                f"{{{self.ns.xsi}}}schemaLocation": self.ns.schema_locations(),
            },
            nsmap=self.ns.nsmap(),
        )

        if len(self.waypoints) > 1:
            waypoints_table = Element(f"{{{self.ns.fpl}}}waypoint-table")
            for waypoint in self.waypoints:
                waypoints_table.append(waypoint.encode())
            root.append(waypoints_table)

        if self.route is not None:
            root.append(self.route.encode())

        document = ElementTree(root)
        return element_string(document, pretty_print=True, xml_declaration=True, encoding="utf-8")

    def dump_xml(self, path: Path):
        with open(path, mode="w") as xml_file:
            xml_file.write(self.dumps_xml().decode())

    def validate(self) -> None:
        """
        Validates the contents of a record against a XSD schema

        The external `xmllint` binary is used to validate records as the `lxml` methods did not easily support relative
        paths for schemas that use imports/includes..

        Schemas are loaded from an XSD directory within this package using the `importlib.files` method. The current
        flight plan object is written to a temporary directory to pass to the `xmllint` binary.

        The `xmllint` binary only returns a 0 exit code if the record validates successfully. Therefore, any other exit
        code can be considered a validation failure, and returned as a `RecordValidationError` exception.
        """
        with TemporaryDirectory() as document_path:
            document_path = Path(document_path).joinpath("fpl.xml")
            self.dump_xml(path=document_path)

            try:
                # Exempting Bandit/flake8 security issue (using subprocess)
                # It is assumed that there are other protections in place to prevent untrusted input being a concern.
                # Namely that this is package will be ran in a secure/controlled environments against pre-trusted files.
                #
                # Use `capture_output=True` in future when we can use Python 3.7+
                subprocess.run(  # noqa: S274,S603 - nosec
                    args=["xmllint", "--noout", "--schema", str(Fpl.schema_path), str(document_path)],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    check=True,
                )
            except subprocess.CalledProcessError as e:
                raise RuntimeError(f"Record validation failed: {e.stderr.decode()}") from e



from datetime import datetime
from typing import List, Dict, Union


def _convert_coordinate_dd_2_ddm(coordinate: float, positive_symbol: str, negative_symbol: str) -> Dict[str, float]:
    coordinate_elements: List[str] = str(coordinate).split(".")

    ddm_coordinate: Dict[str, Union[float, str]] = {
        "degree": abs(float(coordinate_elements[0])),
        "minutes": 0,
        "sign": negative_symbol,
    }

    try:
        ddm_coordinate["minutes"] = abs(float(f"0.{coordinate_elements[1]}") * 60.0)
    except IndexError:
        pass

    if coordinate >= 0:
        ddm_coordinate["sign"] = positive_symbol

    return ddm_coordinate


def convert_coordinate_dd_2_ddm(lon: float, lat: float) -> Dict[str, str]:
    lon = _convert_coordinate_dd_2_ddm(lon, positive_symbol="E", negative_symbol="W")
    lat = _convert_coordinate_dd_2_ddm(lat, positive_symbol="N", negative_symbol="S")

    return {
        "lon": f"{int(lon['degree'])}° {'{:.6f}'.format(lon['minutes'])}' {lon['sign']}",
        "lat": f"{int(lat['degree'])}° {'{:.6f}'.format(lat['minutes'])}' {lat['sign']}",
    }


def file_name_with_date(name: str) -> str:
    return name.replace("{{date}}", f"{datetime.utcnow().date().strftime('%Y_%m_%d')}")



from importlib.metadata import version
from pathlib import Path

import click

from bas_air_unit_network_dataset import NetworkManager


class AppCommand(click.core.Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params.insert(
            0,
            click.core.Option(
                (
                    "-d",
                    "--dataset-path",
                ),
                help="Path to network dataset",
                required=True,
                envvar="AIRNET_DATASET_PATH",
                type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True),
            ),
        )


def inspect_network(network: NetworkManager) -> None:
    click.echo(network)
    click.echo("")

    click.echo(f"Waypoints [{len(network.waypoints)}]:")
    for i, waypoint in enumerate(network.waypoints):
        click.echo(f"{str(i + 1).zfill(2)}. {waypoint}")
    click.echo("")

    click.echo(f"Routes [{len(network.routes)}]:")
    for i, route in enumerate(network.routes):
        click.echo(f"{str(i + 1).zfill(2)}. {route}")


@click.group()
@click.version_option(version("bas_air_unit_network_dataset"))
def cli():
    """BAS Air Unit Network Dataset (`airnet`)"""
    pass


@cli.command()
@click.option(
    "-d",
    "--dataset-path",
    help="Path to network dataset directory",
    required=True,
    envvar="AIRNET_DATASET_PATH",
    type=click.Path(exists=False, file_okay=False, dir_okay=True, readable=True, writable=True, resolve_path=True),
)
def init(dataset_path: str) -> None:
    """Initialises an empty network"""
    _dataset_path = Path(dataset_path).joinpath("bas-air-unit-network-dataset.gpkg")
    click.echo(f"Dataset will be located at: {_dataset_path}")
    click.echo("")

    network = NetworkManager(dataset_path=_dataset_path, init=True)
    click.echo(f"Dataset created at: {network.dataset_path.absolute()}")


@cli.command(cls=AppCommand, name="import")
@click.option(
    "-i",
    "--input-path",
    help="Path to input file containing routes and waypoints",
    required=True,
    envvar="AIRNET_INPUT_PATH",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True),
)
def _import(dataset_path: str, input_path: str) -> None:
    """Import routes and waypoints into network from an input file"""
    _dataset_path = Path(dataset_path)
    _input_path = Path(input_path)
    click.echo(f"Dataset is located at: {_dataset_path}")
    click.echo(f"Input is located at: {_input_path}")
    click.echo("")

    # `init` is set to reset contents of dataset
    network = NetworkManager(dataset_path=Path(dataset_path), init=True)
    network.load_gpx(path=_input_path)
    inspect_network(network=network)

    click.echo("")
    click.echo("Import complete")


@cli.command(cls=AppCommand)
@click.option(
    "-o",
    "--output-path",
    help="Path to save outputs",
    required=True,
    envvar="AIRNET_OUTPUT_PATH",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, writable=True, resolve_path=True),
)
def export(dataset_path: str, output_path: str) -> None:
    """Export network, routes, waypoints as CSV/GPX/FPL outputs"""
    _dataset_path = Path(dataset_path)
    _output_path = Path(output_path)
    click.echo(f"Dataset is located at: {_dataset_path}")
    click.echo(f"Output directory is: {_output_path}")
    click.echo("")

    network = NetworkManager(dataset_path=_dataset_path, output_path=_output_path)

    network.dump_csv()
    print("- CSV export complete")
    network.dump_gpx()
    print("- GPX export complete")
    network.dump_fpl()
    print("- FPL export complete")

    click.echo("")
    click.echo("Export complete")


@cli.command(cls=AppCommand)
def inspect(dataset_path: str) -> None:
    """Inspect state of network, routes, waypoints"""
    _dataset_path = Path(dataset_path)
    click.echo(f"Dataset is located at: {_dataset_path}")
    click.echo("")

    network = NetworkManager(dataset_path=Path(dataset_path))
    inspect_network(network=network)


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bas_air_unit_network_dataset',
 'bas_air_unit_network_dataset.exporters',
 'bas_air_unit_network_dataset.schemas',
 'bas_air_unit_network_dataset.schemas.garmin']

package_data = \
{'': ['*']}

install_requires = \
['Fiona>=1.8.21,<2.0.0',
 'Shapely>=1.8.2,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'gpxpy>=1.5.0,<2.0.0',
 'lxml>=4.9.1,<5.0.0',
 'ulid-py>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['airnet = bas_air_unit_network_dataset.__main__:cli']}

setup_kwargs = {
    'name': 'bas-air-unit-network-dataset',
    'version': '0.2.2',
    'description': 'Utility to process routes and waypoints used by the British Antarctic Survey (BAS) Air Unit',
    'long_description': '# BAS Air Unit Network Dataset\n\nManagement of the network of routes and waypoints used by the British Antarctic Survey (BAS) Air Unit.\n\nIncluding a utility for the Air Unit to process routes and waypoints for use in their handheld and aircraft GPS devices.\n\n## Overview\n\n### Purpose\n\nTo support the BAS Air Unit manage their network of routes and waypoints such that:\n\n* information is internally consistent, through defined structures and constraints\n* information is interoperable between different systems, through the use of open/standard formats\n* information is well described and sharable with other teams, through distribution as datasets\n\n### Background\n\nThis project was developed in response to discussions and requests with the BAS Air Unit to review and simplify the \nprocess they used to manage their network of waypoints, and to ensure it\'s future sustainability.\n\nBAS staff can read more about this background in this \n[GitLab issue](https://gitlab.data.bas.ac.uk/MAGIC/ops-support/-/issues/134) [Internal].\n\n### Status\n\nThis project is an early alpha, working towards a Minimal Viable Product (MVP). This means:\n\n* all, or parts, of this project: \n  * may not be working \n  * may stop working at any time\n  * may not work correctly, or as expectedly (including destructively)\n  * may change at any time (in terms of implementation or functionality)\n* documentation may be missing or incorrect\n* no support is provided for this project\n* no real and/or sensitive information should be used with this project\n\n### Limitations\n\nThis service has a number of limitations, including:\n\n* the default base map used in Garmin BaseCamp is not useful for Antarctica\n* the Air Unit Network utility is only supported on Windows computers\n* the Air Unit Network utility does not support multiple users importing data at the same time\n* the Air Unit Network utility does not create backups when importing new data\n* the Air Unit Network utility does not require route names to follow the required naming convention\n* the Air Unit Network utility does not require waypoint identifiers to be unique across all waypoints\n* the Air Unit Network utility does not require waypoint comments to follow the required structure\n* the Air Unit Network utility does not require waypoints within imported routes to be listed as standalone waypoints\n* contextual comments for waypoints within a route (e.g. \'Start of route\') cannot be set using Garmin BaseCamp\n* comments for waypoints use an overly complex structure to support an ad-hoc serialisation format\n* Unicode characters (such as emoji) cannot be used in route/waypoint names, comments, etc.\n* CSV outputs are not designed for printing (i.e. column formatting and page breaks)\n* the workspace directory and its data is only available at Rothera (i.e. it isn\'t synced back to Cambridge)\n\nSome or all of these limitations will be addressed in [Future improvements](#future-improvements) to this project.\n\n### Future improvements\n\nSee the project [issue tracker](#issue-tracking) for a backlog of planned features and ideas to further improve this\nproject.\n\n## Usage\n\n### Workspace directory\n\nThe *workspace directory* contains all files related to this project. This directory should be shared between all \nusers/computers that require access by storing on a shared drive or synced folder accessible to relevant users.\n\n**Note:** To create a new workspace directory, see the [Setup](#setup) section.\n\nA typical/example workspace directory contains:\n\n```\n/path/to/workspace/directory\n├── bas-air-unit-network-dataset.gpkg\n├── input.gpx\n└── output/\n    ├── CSV/\n    │   ├── 00_WAYPOINTS_2022_07_08.csv\n    │   ├── 00_WAYPOINTS_2022_07_08_DD.csv\n    ├── FPL/\n    │   ├── 00_WAYPOINTS_2022_07_08.fpl\n    │   ├── 01_BRAVO_TO_ALPHA.fpl\n    │   ├── 02_BRAVO_TO_BRAVO.fpl\n    │   └── 03_bBRAVO_TO_LIMA.fpl\n    └── GPX/\n        └── 00_NETWORK_2022_07_08.gpx\n```\n\n* the `output/` directory contains files for use in GPS devices and as print outs, organised by file type (CSV, GPX \n  and FPL)\n* the `input.gpx` file contains routes and waypoints *exported* from Garmin BaseCamp to be *imported* into the Air \n  Unit Network utility\n* the `bas-air-unit-network-dataset.gpkg` file is a GeoPackage used internally by the Air Unit Network utility\n\n#### Access control\n\nThe Air Unit Network utility does not include access control, therefore the permissions applied to the working directory\nshould be used if access to information needs to be restricted. It is recommended that wrtite permissions on the \ndirectory are restricted to users that need to edit information.\n\n### Workflow\n\n**Note:** You need to complete the steps in the [Installation](#installation) section to complete this workflow.\n\nTo update waypoints and routes (the network) and create new files for use in GPS devices and as print outs:\n\n1. [waypoints](#managing-waypoints-in-basecamp) and [routes](#managing-routes-in-basecamp) are created and edited using \n   Garmin BaseCamp\n2. [export routes and waypoints from BaseCamp](#saving-waypoints-and-routes-from-basecamp) as a GPX file\n3. [import the GPX file into the Air Unit Network utility](#importing-waypoints-and-routes-into-the-network-utility)\n4. use the utility to [export the network in a range of formats](#exporting-waypoints-and-routes-using-the-network-utility) \n   (CSV, GPX and FPL)\n\nThe sub-sections below describe these steps in more detail.\n\n### Managing waypoints in BaseCamp\n\n#### Create a new waypoint\n\n1. from the start menu, launch *BaseCamp*\n2. from the upper left hand panel, under *My Collection*, select *BAS Air Unit Network* -> *BAS Air Unit Network*\n3. choose the *Create a waypoint* tool (flag icon) from the top middle toolbar (creation tools)\n4. select the general location of the waypoint (precise position can be entered later)\n5. from the *General* tab of the waypoint properties screen that opens:\n   1. enter a suitable *Name* or identifier (maximum 6 characters)\n   2. enter a suitable *Comment*, which consists of 5 elements in the order below, each separated with a `|` (vertical \n      bar) character:\n      1. *name*: a full, or formal name for the waypoint (maximum 17 characters)\n      2. *co-located with*: name of a related depot, instrument and/or other feature - use `N/A` if not relevant\n      3. *last accessed at*: date waypoint was last accessed in the form `YYYY-MM-DD` - use `N/A` if unvisited\n      4. *last accessed by*: pilot that that last accessed waypoint - use `N/A` if unvisited\n      5. *other information*: any other information - use `N/A` if not relevant\n6. from the *Advanced* tab:\n   1. if necessary, edit the *Position* to the correct latitude and longitude in degrees, decimal minutes (DDM) format\n7. close the waypoint properties screen\n\nFor example (a co-located, previously visited, waypoint with a full name and additional information):\n\n* name: `ALPHA`\n* comment: `Alpha 001 | Dog | 2014-12-24 | CW | Bring treats.`\n* position: `S69° 54.910\' W75° 00.878\'`\n\nFor example (a standalone, unvisited, waypoint with no full/formal name or additional information):\n\n* name: `BRAVO`\n* comment: `N/A | N/A | N/A | N/A | N/A`\n* position: `S70° 49.810\' W75° 11.420\'`\n\n**Note:** Only the \'name\' in a comment will be included in FPL waypoints.\n\n#### Edit an existing waypoint\n\n1. from the start menu, launch *BaseCamp*\n2. from the upper left hand panel, under *My Collection*, select *BAS Air Unit Network* -> *BAS Air Unit Network*\n3. from the lower left hand panel, right click the waypoint to be edited and select *Get Info*\n4. from the waypoint properties screen, update (as needed):\n   1. the name (identifier)\n   2. comment (see rules above for how comments must be written)\n   3. position\n\n### Managing routes in BaseCamp\n\n#### Create a new route\n\n1. from the start menu, launch *BaseCamp*\n2. from the upper left hand panel, under *My Collection*, select *BAS Air Unit Network* -> *BAS Air Unit Network*\n3. choose the *Create a route* tool (path/network icon) from the top middle toolbar (creation tools)\n4. from the window that opens:\n   1. enter the *starting* waypoint using the search box\n   2. enter the *ending* waypoint using the search box\n   3. select the *Go* command\n5. from the *Via Points* tab of the route properties screen that opens:\n   1. un-tick the *Autoname* checkbox\n   2. rename the route as per the route naming convention: \n       * `{Route Number}_{Start Waypoint Identifier}_To_{End Waypoint Identifier}`\n       * for example: `01_ALPHA_TO_BRAVO`\n       * the first route should use `01` as a route number, `00` is a reserved value and should not be used\n   3. if additional waypoints should be added between start and end waypoints:\n      1. from the lower left navigation panel, drag waypoints into the route properties screen, in its position within \n         the route\n   4. close the route properties screen\n\n#### Edit an existing route\n\n1. from the start menu, launch *BaseCamp*\n2. from the upper left hand panel, under *My Collection*, select *BAS Air Unit Network* -> *BAS Air Unit Network*\n3. from the lower left hand panel, right click the route to be edited and select *Get Info*\n4. to remove a waypoint from the route:\n   1. right click the route waypoint from the list and select *Remove*\n5. to add a new waypoint to the route:\n   1. from the lower left navigation panel, drag waypoints into the route properties screen, in its position within \n      the route\n6. if the start or end waypoint in the route has changed, update the route name (see rules above for format)\n\n### Saving waypoints and routes from BaseCamp \n\n1. from the start menu, launch *BaseCamp*\n2. from the upper left hand panel, under *My Collection*, select *BAS Air Unit Network* -> *BAS Air Unit Network*\n3. from the *File* menu, select *Export "BAS Air Unit Network"*\n4. name the file `input.gpx` and save to the [Workspace directory](#workspace-directory) (the existing file should be \n   overwritten)\n5. set the export format as *GPX v1.1*\n\n### Importing waypoints and routes into the Network utility\n\n1. from the start menu, launch *Command Prompt* and run the commands in [1]\n\n**WARNING:** Importing replaces all existing information. Ensure you have suitable backups of existing data.\n\n[1]\n\n**Note:** If viewing this documentation as a PDF, some commands below may disappear off the page. If you select the \nline you should be able to copy into a tool such as Notepad to view it in full.\n\n```\n$ C:\\ProgramData\\Miniconda3\\envs\\airnet\\Scripts\\activate.bat\n\n# navigate to the workspace directory\n(airnet) $ cd \'/path/to/workspace/directory\'\n\n(airnet) $ C:\\ProgramData\\Miniconda3\\envs\\airnet\\python.exe C:\\ProgramData\\Miniconda3\\envs\\airnet\\Scripts\\airnet.exe import --dataset-path bas-air-unit-network-dataset.gpkg --input-path input.gpx\nDataset is located at: /path/to/workspace/directory/bas-air-unit-network-dataset.gpkg\nInput is located at: /path/to/workspace/directory/input.gpx\n\n<NetworkManager : 12 Waypoints - 3 Routes>\n\nWaypoints [12]:\n01. <Waypoint 01G7T144RF7WENM8PBKC15W1YB :- [ALPHA_], POINT (-75.01463335007429 -69.91516669280827)>\n02. <Waypoint 01G7T144RFVMA8XSD2NES8EFW0 :- [BRAVO_], POINT (-75.19033329561353 -70.8301666751504)>\n03. <Waypoint 01G7T144RFBM65YQC317D7XCC8 :- [CHARLE], POINT (-71.49899996817112 -71.76016665995121)>\n04. <Waypoint 01G7T144RFXNTDVKXCEBTW4DSJ :- [DELTA_], POINT (-66.09375 -72.7509999834001)>\n05. <Waypoint 01G7T144RFMSAZZQEAD5ANBTAN :- [ECHO__], POINT (-61.611166689544916 -73.45333331264555)>\n06. <Waypoint 01G7T144RFVJFRW1VY0MFQNW06 :- [FOXTRT], POINT (-70.22449999116361 -73.17583330906928)>\n07. <Waypoint 01G7T144RFSM18AJ9G2A3GKGZ9 :- [GOLF__], POINT (-70.3125 -74.20000003650784)>\n08. <Waypoint 01G7T144RFFA3B2009P6FAXC3H :- [HOTEL_], POINT (-65.43449998833239 -74.42566668614745)>\n09. <Waypoint 01G7T144RFP1F3EB58V421Y65S :- [INDIA_], POINT (-63.017566641792655 -75.70466665551066)>\n10. <Waypoint 01G7T144RFTW3RN38TNR6TJ585 :- [JULIET], POINT (-58.93050002865493 -76.34150002151728)>\n11. <Waypoint 01G7T144RFV5N138VMTC80WTA1 :- [KILO__], POINT (-55.37100004032254 -77.16683333739638)>\n12. <Waypoint 01G7T144RFNDB0TEMY911J8VN7 :- [LIMA__], POINT (-50.8461666572839 -78.08921668678522)>\n\nRoutes [3]:\n01. <Route 01G7T144RF040ETQSEBGS73A9Z :- [01_BRAVO_TO_ALPHA], 2 waypoints, Start/End: BRAVO  / ALPHA >\n02. <Route 01G7T144RFAN8RKG006NJ2CC4B :- [02_BRAVO_TO_BRAVO], 9 waypoints, Start/End: BRAVO  / BRAVO >\n03. <Route 01G7T144RFGTXKKJGA1V6CM24F :- [03_BRAVO_TO_LIMA], 8 waypoints, Start/End: BRAVO  / LIMA  >\n\nImport complete\n```\n\n### Exporting waypoints and routes using the Network utility\n\n1. from the start menu, launch *Command Prompt* and run the commands in [1]\n\n**Note:** If you encounter an error similar to [2], ensure you install the \n[Microsoft Visual C++ Redistributable](#microsoft-visual-c-redistributable) from the installation bundle.\n\n[1]\n\n**Note:** If viewing this documentation as a PDF, some commands below may disappear off the page. If you select the \nline you should be able to copy into a tool such as Notepad to view it in full.\n\n```\n$ C:\\ProgramData\\Miniconda3\\envs\\airnet\\Scripts\\activate.bat\n\n# navigate to the workspace directory\n(airnet) $ cd \'/path/to/workspace/directory\'\n\n(airnet) $ C:\\ProgramData\\Miniconda3\\envs\\airnet\\python.exe C:\\ProgramData\\Miniconda3\\envs\\airnet\\Scripts\\airnet.exe export --dataset-path bas-air-unit-network-dataset.gpkg --output-path output/\nDataset is located at: /path/to/workspace/directory/bas-air-unit-network-dataset.gpkg\nOutput directory is is: /path/to/workspace/directory/output\n\n- CSV export complete\n- GPX export complete\n- FPL export complete\n\nExport complete\n```\n\n[2]\n\n> The code execution cannot proceed because VCRUNTIME140.dll was not found. Reinstalling the program may fix this \n> problem.\n\n## Installation\n\n### Install software\n\n**Note:** You will need an internet connection, administrator rights and access to the \n[Installation bundle](#installation-bundle) to install the Air Unit Network utility.\n\nTo install software needed to manage the Air Unit Network dataset:\n\n1. ensure all required Windows updates are installed\n2. download, or copy from a hard drive, the [Installation bundle](#installation-bundle)\n   1. these instructions assume you will download, or copy, the bundle to your *Downloads* directory\n3. from the installation bundle, install Garmin BaseCamp, and optionally, QGIS:\n   1. run `garmin-basecamp-installer.exe` to install Garmin BaseCamp\n   2. optionally, run `qgis-installer.msi` to install the Long Term Support (LTS) version of QGIS\n4. from the installation bundle, install Miniconda (needed to run the Air Unit Network utility):\n   1. run `miniconda-installer.exe` to install Miniconda\n   2. select the *All users (requires admin privileges)* installation option when asked \n5. from the installation bundle, install LibXML2 (needed to run the Air Unit Network utility):\n   1. unzip the `libxml2.zip` archive to a temporary directory\n   2. copy the `libxml2` directory to `C:\\Program Files`, such that `C:\\Program Files\\libxml2\\bin\\xmllint.exe` exists\n   3. from Windows Explorer, right click *This PC* from the left hand panel and click *Properties* \n      1. from the left hand menu of the System Control Panel screen, click *Advanced system settings*\n      2. from the *Advanced* tab of the System Properties screen, click *Environment Variables*\n      3. from the bottom section (System variables) select the *Path* variable and click *Edit*:\n         1. from the Edit environment variable screen, click *New*\n         2. enter `C:\\Program Files\\libxml2\\bin` as a new entry (at the bottom of the list)\n      4. click *OK*, then *OK*, then *OK* again and close Control Panel\n   4. delete the temporary directory created earlier\n6. from the installation bundle, install the Air Unit Network utility virtual environment:\n   1. unzip the `airnet-virtual-environment.zip` archive to a temporary directory (this will take some time)\n   2. from the *View* tab in Explorer, check the *Hidden items* checkbox\n   3. copy the `airnet` directory to `C:\\ProgramData\\Miniconda3\\envs`, such that \n   `C:\\ProgramData\\Miniconda3\\envs\\airnet\\Scripts\\airnet.exe` exists \n   4. delete the temporary directory created earlier\n   5. optionally, uncheck the *Hidden items* checkbox in Explorer\n\n**Note:** Only versions of software within the Installation Bundle are supported.\n\n#### Microsoft Visual C++ redistributable\n\nIn some cases you also need to install the Microsoft Visual C++ Redistributable package, which is needed for LibXML2.\n\n**Note:** Usually a version of this package will already be installed and these instructions are not necessary.\n\nTo install the package:\n\n1. from the installation bundle run `microsoft-vc-installer.exe`\n2. reboot the computer after installation\n\n### Configure software\n\n#### Configure Garmin BaseCamp\n\nTo configure Garmin BaseCamp needed to manage the Air Unit Network dataset:\n\n1. from the start menu, launch *BaseCamp*\n2. when asked for an activity type, select *Direct*\n3. from the upper left hand panel, under *My Collection*:\n   1. right-click and select *New List folder*\n   2. name the folder `BAS Air Unit Network`\n4. within this new folder:\n   1. right-click and select *New List*\n   2. name the list `BAS Air Unit Network`\n5. from the upper left hand panel, under *My Collection*, select *BAS Air Unit Network* -> *BAS Air Unit Network*\n6. from the *File* menu, select *Import into "BAS Air Unit Network"*\n7. select the `input.gpx` file from the [Workspace directory](#workspace-directory)\n\n### Installation bundle\n\nThe installation bundle is a folder of supported software installers and packages. It is designed to run offline and be\ndistributed via an external hard-drive.\n\nThe definitive installation bundle is stored in the\n[MAGIC Office 365 OneDrive](https://nercacuk.sharepoint.com/:f:/s/BASMagicTeam/EvFtTOCBeClCgflcEju58OEBc5xeU0LuTxjxUwQ_V55LVg?e=Fs1gAi)\nand is accessible to all BAS staff. When South, MAGIC will hold a copy of the installation bundle on a hard drive.\n\nSoftware in the installation bundle will be updated periodically after testing, and when new versions of the Air Unit \nNetwork utility are released.\n\n**Note:** Only versions of software within the Installation Bundle are supported.\n\n## Implementation\n\nThis project consists of:\n\n* a description of the waypoints and routes datasets for the BAS Air Unit\n* a Python command line utility to:\n  * store waypoints and routes in a structured/neutral format (currently a GeoPackage)\n  * import waypoints and routes from a GPX file from an editor such as Garmin BaseCamp\n  * export waypoints and routes into a range of output formats (currently CSV, GPX and Garmin FPL)\n\n### Information model\n\nThe BAS Air Unit Network information model consists of two entities, forming two, related, datasets:\n\n1. **Waypoints**: Features representing landing sites used by the Air Unit, usually co-located with a BAS Operations \n   depot, field camp or a science/monitoring instrument \n2. **Routes**: Features representing formally defined, frequently travelled, paths between two or more Waypoints, as \n   opposed to ad-hoc paths\n\nFor example:\n\n* **Waypoints**: Fossil Bluff\n* **Routes**: Rothera to Fossil Bluff\n\nThere is a many-to-many relationship between Waypoints and Routes. I.e. a Waypoint can be part of many Routes, and \nRoutes can contain many Waypoints.\n\n**Note:** This information model is abstract and requires implementing. See the [Data model](#data-model) section for \nthe current implementation.\n\n#### Waypoints (information model)\n\n| Property           | Name             | Type                              | Occurrence | Length | Description                                                | Example                                   |\n|--------------------|------------------|-----------------------------------|------------|--------|------------------------------------------------------------|-------------------------------------------|\n| `id`               | ID               | String                            | 1          | 1 - .. | Unique identifier                                          | \'01G7MY680N332AW9H9HR9SG15T\'              |\n| `identifier`       | Identifier       | String                            | 1          | 1 - 6  | Unique reference                                           | \'ALPHA\'                                   |\n| `geometry`         | Geometry         | Geometry (2D/3D Point, EPSG:4326) | 1          | -      | Position or location as a single coordinate                | \'SRID=4326;Point(-75.014648 -69.915214)\'  |\n| `name`             | Name             | String                            | 0-1        | 1 - 17 | Full or formal name                                        | \'Alpha 001\'                               |\n| `colocated_with`   | Co-located With  | String                            | 0-1        | 1 - .. | Features (from other domains) associated with the waypoint | \'Depot: Foo\'                              |\n| `last_accessed_at` | Last Accessed At | Date                              | 0-1        | 1 - .. | When the Waypoint was last accessed or visited             | \'2014-12-24\'                              |\n| `last_accessed_by` | Last Accessed By | String                            | 0-1        | 1 - .. | Who last accessed or visited the Waypoint                  | \'Conwat\'                                  |                            \n| `comment`          | Comment          | String                            | 0-1        | 1 - .. | Freetext description or comments                           | \'Alpha 001 is on a high ridge ...\'        |\n\n##### ID (Waypoint)\n\nIDs:\n\n* MUST be unique\n* MUST NOT be based on any information contained within the Waypoint\n* MAY use any format/scheme:\n  * the same scheme SHOULD be used for all IDs\n  * non-sequential schemes are recommended\n\n**Note:** This ID can be used to refer to each Waypoint in other systems (i.e. as an foreign identifier).\n\n##### Identifiers (Waypoint)\n\nIdentifiers:\n\n* MUST be between 1 and 6 uppercase alpha-numeric characters without spaces (A-Z, 0-9)\n* MUST be unique across all Waypoints\n\n##### Geometry (Waypoint)\n\nGeometries:\n\n* MUST be expressed in decimal degrees using the EPSG:4326 projection.\n* MUST consist of either:\n  * a longitude (X) and latitude (Y) dimension (2D point)\n  * a longitude (X), latitude (Y) and elevation (Z) dimension (3D point)\n\n##### Name (Waypoint)\n\nIf specified:\n\n* MUST be between 1 and 17 uppercase alpha-numeric or space characters (A-Z, 0-9, \' \')\n\n##### Co-located with (Waypoint)\n\nNo special comments.\n\n##### Last accessed at (Waypoint)\n\nIf specified:\n\n* MUST be expressed as an [ISO 8601-1:2019](https://www.iso.org/standard/70907.html) date instant\n\n##### Last accessed by (Waypoint)\n\nIf specified:\n\n* MUST unambiguously reference an individual\n* MAY use any scheme:\n  * the same scheme SHOULD be used for all Waypoints\n\n##### Comment (Waypoint)\n\nNo special comments.\n\n#### Routes (information model)\n\n| Property          | Name      | Type                      | Occurrence | Length | Description           | Example                      |\n|-------------------|-----------|---------------------------|------------|--------|-----------------------|------------------------------|\n| `id`              | ID        | String                    | 1          | 1 - .. | Unique identifier     | \'01G7MZB9X0R8S7RTNYAMAQKHE4\' |\n| `name`            | Name      | String                    | 1          | 1 - .. | Name or reference     | \'01_ALPHA_TO_BRAVO\'          |\n| `waypoints`       | Waypoints | List of Waypoint entities | 2-n        | -      | Sequence of Waypoints | -                            |\n\n##### ID (Route)\n\nIDs:\n\n* MUST be unique\n* MUST NOT be based on any information contained within the Route\n* MAY use any format/scheme:\n  * the same scheme SHOULD be used for all IDs\n  * non-sequential schemes are recommended\n\n**Note:** This ID can be used to refer to each Route in other systems (i.e. as an foreign identifier).\n\n##### Name (Route)\n\nNames:\n\n* MUST use the format `{Sequence}_{First Waypoint Identifier}_TO_{Last Waypoint Identifier}`, where `{Sequence}` is \n  a zero padded, auto-incrementing prefix (e.g. \'01_\', \'02_\', ..., \'99_\').\n\n##### Waypoints (Route)\n\nWaypoints in Routes:\n\n* MUST be a subset of the set of Waypoints\n  * i.e. waypoints in routes MUST be drawn from a common set, rather than defined ad-hoc or inline within a Route\n* MUST be expressed as a sequence:\n  * i.e. a list in a specific order from a start to finish via any number of other places\n* MUST contain at least 2 Waypoints but MAY contain more:\n  * i.e. a start and end\n* MAY be included multiple times\n  * i.e. the start and end can be the same Waypoint, or a route may pass through the same waypoint multiple times\n* MAY include a contextual comment or description\n  * e.g. \'Start of route\'\n\n### Data model\n\nThe BAS Air Unit Network data model implements the [Information model](#information-model) using three entities:\n\n1. **Waypoint**: Point features with attributes\n2. **Route**: Features to contextualise a set of Waypoints, with attributes (such as route name)\n3. **RouteWaypoint**: join between a Waypoint and a Route, with contextual attributes (such as sequence within route)\n\nThis data model describes how these entities are:\n\n* persisted as features within layers within an OGC GeoPackage (version 1.2)\n* represented as Python class instances (objects)\n\nThis GeoPackage, and the data it contains, is considered the Source of Truth and definitive version/format.\n\n**Note:** This data model does not describe how entities are encoded in specific [Output Formats](#output-formats).\n\n**Note:** This data model does not describe specifics about how entities are encoded in GeoPackages (i.e. database \nschema), or describe GeoPackage meta tables, which are managed automatically by libraries used by this utility.\n\n#### FIDs\n\nFeature Identifiers (FIDs) are created automatically when features are persisted to GeoPackages. FIDs are unique \nauto-incrementing integers, and used as primary keys within database tables.\n\nFIDs SHOULD be considered an implementation detail, and SHOULD be ignored in favour of ID properties (i.e. \'ID\' rather \nthan \'FID\').\n\nFIDs SHOULD NOT be exposed to end users and their values or structure MUST NOT be relied upon.\n\n#### ULIDs\n\n[Universally Unique Lexicographically Sortable Identifier (ULID)](https://github.com/ulid/spec)s are the scheme used \nfor identifiers (IDs).\n\nThese IDs MAY be exposed to end users.\n\n#### Waypoints (data model)\n\nPython class: \n\n* `Waypoint` (single waypoint)\n* `WaypointCollection` (waypoints set)\n\nGeoPackage layer: `waypoints`\n\n| Property           | Name             | Data Type     | Nullable | Unique | Max Length | Notes                                                |\n|--------------------|------------------|---------------|----------|--------|------------|------------------------------------------------------|\n| `fid`              | Feature ID       | Integer       | No       | Yes    | -          | Internal to database, primary key, auto-incrementing |\n| `id`               | ID               | ULID (String) | No       | Yes    | -          | -                                                    |\n| `identifer`        | Identifier       | String        | No       | Yes    | 6          | -                                                    |\n| `geometry`         | Geometry         | 2D/3D Point   | No       | No     | -          | -                                                    |\n| `name`             | Name             | String        | Yes      | No     | 17         | -                                                    |\n| `colocated_with`   | Co-located With  | String        | Yes      | No     | -          | -                                                    |\n| `last_accessed_at` | Last Accessed At | Date          | Yes      | No     | -          | -                                                    |\n| `last_accessed_by` | Last Accessed By | String        | Yes      | No     | -          | -                                                    |\n| `comment`          | Comment          | String        | Yes      | No     | -          | -                                                    |\n\n#### Routes (data model)\n\nPython class: \n\n* `Route` (single route)\n* `RouteCollection` (routes set)\n\nGeoPackage layer: `routes`\n\n| Property | Name       | Data Type       | Nullable | Unique | Max Length | Notes                                                |\n|----------|------------|-----------------|----------|--------|------------|------------------------------------------------------|\n| `fid`    | Feature ID | Integer         | No       | Yes    | -          | Internal to database, primary key, auto-incrementing |\n| `id`     | ID         | ULID (String)   | No       | Yes    | -          | -                                                    |\n| `name`   | Name       | String          | No       | Yes    | -          | -                                                    |\n\n#### Route Waypoints (data model)\n\nPython class: \n\n* `RouteWaypoint` (single waypoint in route)\n\nGeoPackage layer: `route_waypoints`\n\n| Property      | Name        | Data Type      | Nullable | Unique             | Max Length | Notes                                                                       |\n|---------------|-------------|----------------|----------|--------------------|------------|-----------------------------------------------------------------------------|\n| `fid`         | Feature ID  | Integer        | No       | Yes                | -          | Internal to database, primary key, auto-incrementing                        |\n| `route_id`    | Route ID    | ULID (String)  | No       | Yes                | -          | Foreign key to Route entity                                                 |\n| `waypoint_id` | Waypoint ID | ULID (String)  | No       | Yes                | -          | Foreign key to Waypoint entity                                              |\n| `sequence`    | Sequence    | Integer        | No       | Yes (within Route) | -          | Position of waypoint within a route, value must be unique within each route |\n| `comment`     | Comment     | String         | Yes      | No                 | -          | -                                                                           |\n\n**Note:** Though the `route_id` and `waypoint_id` columns are effectively foreign keys, they are not configured as \nsuch within the GeoPackage.\n\n### Source data\n\nInput/source data MUST be saved in a GPX (GPS Exchange Format) version 1.1 file to be imported into this project.\n\n**Note:** Only GPX files produced by Garmin BaseCamp are supported. GPX files produced by other tools may also work but \nare not formally supported.\n\nLimitations:\n\n* Garmin BaseCamp does not support custom GPX extensions, except Garmin\'s own, limiting fields to the core GPX \n  specification\n* this requires non-standard fields (such as `waypoint.last_accessed_at`) to be added to the comment freetext field, \n  making it more complex (see notes in the [Creating a Waypoint](#create-a-new-waypoint) section)\n\n### Test network\n\nA test network of 12 waypoints and 3 routes is used to:\n\n1. test various edge cases\n2. provide consistency for repeatable testing\n3. prevent needing to use real data that might be sensitive\n\n**Note:** Route and waypoints in the test network are arbitrary and do not reflect any actual features.\n\n**WARNING!** This test network is entirely fictitious. It MUST NOT be used for any real navigation.\n\nA [Workspace Directory](#workspace-directory) for the test network is maintained in the\n[MAGIC Office 365 OneDrive](https://nercacuk.sharepoint.com/:f:/s/BASMagicTeam/EhBAbE0tTDxCt298WEPOWoMBtuV7yxOYuJ8bPslVdKlASQ)\nand is accessible to all BAS staff. When South, MAGIC will hold a copy of the test network on a hard drive.\n\nA QGIS project is provided to visualise the test network and ensure exported outputs match expected test data.\n\n### Output formats\n\n#### Supported formats\n\n**Note:** The GeoPackage used in the [Data Model](#data-model) is not a supported output format.\n\nFormat use-cases:\n\n| Format | Use Case                          |\n|--------|-----------------------------------|\n| CSV    | Human readable, printed reference |\n| GPX    | Machine readable, handheld GPS    |\n| FPL    | Machine readable, aircraft GPS    |\n\nFormat details:\n\n| Format | Name                   | Version  | Encoding | Open Format          | Restricted Attributes | Extensions Available | Extensions Used  |\n|--------|------------------------|----------|----------|----------------------|-----------------------|----------------------|------------------|\n| CSV    | Comma Separated Value  | N/A      | Text     | Yes                  | No                    | No                   | N/A              |\n| GPX    | GPS Exchange Format    | 1.1      | XML      | Yes                  | Yes                   | Yes                  | No               |\n| FPL    | (Garmin) Flight Plan   | 1.0      | XML      | No (Vendor Specific) | Yes                   | Yes                  | No               |\n\nOutputs produced for each format: \n\n| Format | Each Waypoint | Each Route | All Waypoints (Only) | All Routes (Only) | Waypoints and Routes (Combined) |\n|--------|---------------|------------|----------------------|-------------------|---------------------------------|\n| CSV    | No            | No         | Yes                  | No [1]            | No                              |\n| GPX    | No            | No         | No [1]               | No [1]            | Yes                             |\n| FPL    | No            | Yes        | Yes                  | No                | No                              |\n\nWhere \'All Waypoints (Only)\' outputs are produced, waypoints will be sorted alphabetically.\n\n[1] These outputs can be produced but are intentionally excluded as they aren\'t used by the Air Unit. See this \n[GitLab issue](https://gitlab.data.bas.ac.uk/MAGIC/air-unit-network-dataset/-/issues/101) [Internal] for details.\n\n#### Output file names\n\nOutput files use an internal naming convention for all formats:\n\n| Export Type                     | File Name (Pattern)                 | File Name (Example)           |\n|---------------------------------|-------------------------------------|-------------------------------|\n| Each Waypoint                   | N/A                                 | N/A                           |\n| Each Route                      | `{route name}.ext`                  | `01_ALPHA_TO_BRAVO.ext`       |\n| All Waypoints (Only)            | `00_WAYPOINTS_{current date}.ext`   | `00_WAYPOINTS_2014_12_24.ext` |\n| All Routes (Only)               | `00_ROUTES_{current date}.ext`      | `00_ROUTES_2014_12_24.ext`    |\n| Waypoints and Routes (Combined) | `00_NETWORK_{current date}.ext`     | `00_NETWORK_2014_12_24.ext`   |\n\nWhere `.ext` is a relevant file extension for each format (i.e. `.csv` for CSV outputs).\n\n#### Output format - CSV\n\nNotes:\n\n* CSV outputs use the first row as a column names header\n* outputs produced for all routes use a `route_name` column to distinguish rows related to each route\n* `waypoint.geometries` can optionally be included as separate latitude (Y) and longitude (X) columns in either:\n  * decimal degrees (`latitude_dd`, `longitude_dd` columns) - native format\n  * degrees, decimal minutes (`latitude_ddm`, `longitude_ddm` columns) - format used in aviation\n\nLimitations:\n\n* all properties are encoded as strings, without type hints using extended CSV schemes etc.\n* `waypoint.geometries` containing an elevation (Z) dimension are not included in CSV outputs\n* CSV outputs are not validated\n\n#### Output format - GPX\n\nNotes:\n\n* GPX outputs are validated against the GPX XSD schema automatically\n\nLimitations:\n\n* GPX metadata fields (author, last updated, etc.) are not currently populated\n* the `waypoint.name`, `waypoint.colocated_with`, `waypoint.last_accessed_at`, `waypoint.last_accessed_by` \n  and `waypoint.comment`, properties are combined into the GPX comment field, as GPX lacks fields for these properties\n* `waypoint.geometries` containing an elevation (Z) dimension are not included in GPX outputs\n\n#### Output format - FPL\n\nNotes:\n\n* FPL outputs are validated against a custom version of the Garmin [FPL XSD schema](#fpl-xml-schema) automatically\n* route names will use spaces instead of underscores in FPL files, as underscores aren\'t allowed in FPL route names\n\nLimitations:\n\n* the `waypoint.colocated_with`, `waypoint.last_accessed_at`, `waypoint.last_accessed_by` and `waypoint.comment` \n  properties are not included in FPL waypoint comments, as they are limited to 17 characters [1]\n* underscores (`_`) characters are stripped from route names *within* FPL files (rather than the names *of* FPL \n  files), a local override is used to replace underscores with spaces (` `) to workaround this limitation\n* FPL metadata fields (author, last updated, etc.) are not currently populated\n* `waypoint.geometries` containing an elevation (Z) dimension are not included in FPL outputs\n\n[1] This limit comes from the specific UI shown in the aircraft GPS used by the BAS Air Unit.\n\n##### FPL XML schema\n\nA copy of the Garmin FPL XML/XSD schema, http://www8.garmin.com/xmlschemas/FlightPlanv1.xsd, is included in this \nproject to locally validate generated FPL outputs. This schema cannot be used for validation in its published form, as\nit contains a number of invalid regular expressions. These regular expressions have been modified in the schema used \nin this project, which hopefully match Garmin\'s intentions.\n\nIn order to produce FPL files that match those produced by earlier processing scripts used by the BAS Air Unit, a \nnumber of other changes have been made to the local version of the FPL schema. These include:\n\n* removing the requirement for a `<waypoints-table>` element to be included in all FPL files (relevant to route FPLs)\n* removing the requirement for all `<waypoint>` elements within `<route>` elements to be included in a \n  `<waypoint-table>` element (as a consequence of the above)\n* altering the regular expression used for the `<country-name>` element to allow the `_` characters\n\n**Note:** It is hoped these local modifications will be removed in future through testing with the in-aircraft GPS.\nSee [#12](https://gitlab.data.bas.ac.uk/MAGIC/air-unit-network-dataset/-/issues/32) for more information.\n\n### Python CLI\n\nA Python based utility, `airnet`, is used for:\n\n* importing [Source data](#source-data) into the internal GeoPackage\n* exporting data from the internal GeoPackage as [Outputs](#output-formats)\n* creating new datasets (setup task, only needs to be performed once)\n* describing the contents of a dataset at a high level\n\nThis utility provides a Command Line Interface (CLI) using the [Click](https://click.palletsprojects.com/en/8.1.x/) \nframework. To enable users to call the CLI from a shell, without needing to run through Python, the Click CLI instance \nis configured as a console script entrypoint in the `setup.py` generated by Poetry (and defined in `pyproject.toml`)\n\nThe Python methods used in the Click CLI call methods from a `NetworkManager` class, which together with a set of \nentity and entity collection classes, is responsible for loading and dumping data.\n\nSee `bas_air_unit_network_dataset.__main__:cli` for the Click CLI method.\n\nSee `bas_air_unit_network_dataset.__init__:NetworkManager` for the NetworkManager and related classes.\n\nFor a list of available commands and their options, run `airnet --help`.\n\nFor information on the installed package version, run `airnet --version`.\n\n### CLI environment variables\n\nParameter values for the `airnet` command line interface (CLI) can be provided using environment variables. These are\ntypically used when automating commands as part of a script or other automation workflow. \n\n| Parameter               | Environment Variable  |\n| ----------------------- | --------------------- |\n| `-d` / `--dataset-path` | `AIRNET_DATASET_PATH` |\n| `-i` / `--input-path`   | `AIRNET_INPUT_PATH`   |\n| `-o` / `--output-path`  | `AIRNET_OUTPUT_PATH`  |\n\n**Note:** Where an option is set on the command line and using an environment variable, the command line value takes\nprecedence.\n\n## Setup\n\n### Populate the installation bundle\n\n**Note:** You will need appropriate rights within the BAS MAGIC team OneDrive to complete these steps.\n\nTo populate the [Installation Bundle](#installation-bundle):\n\n* download the latest Garmin BaseCamp installer (Windows) and rename to `garmin-basecamp-installer.exe`\n* download the latest LTS QGIS installer (Windows) and rename to `qgis-installer.msi`\n* download the latest MiniConda installer (Python 3.9, Windows 64 bit) and rename to `miniconda-installer.exe`\n* download the latest LibXML2 Windows build:\n  * visit the [LibXML2](https://gitlab.gnome.org/GNOME/libxml2/-/tree/master/) Gnome project\n  * view the latest Continuous Deployment pipeline\n  * view the `cmake:msvc` job\n  * download the artefact\n  * extract the artefact Zip, rename the containing directory `libxml2` and rezip the directory as `libxml2.zip`\n* download the \n  [Microsoft Visual C++ 2015 Redistributable Update 3 RC](https://www.microsoft.com/en-us/download/confirmation.aspx?id=52685) \n  package [1] and rename to `microsoft-vc-installer.exe`\n* create a `build` directory and:\n  * download the latest 7-Zip installer and rename to `7zip-installer.exe`\n\nBefore saving installers to the Installation Bundle, test them in a [deployment VM](#setup-a-windows-deployment-vm).\n\n[1] This specific version (14.0.24123) was used because the latest version available at the time of testing (14.32.\n31332, Sept 2022) did not work. In future, it may be that another version is needed, especially if the version of \nLibXML2 changes.\n\n### Setup a workspace directory\n\n1. using Windows Explorer, create a suitable directory to use a Workspace (typically on a shared drive or within a \n   synced folder)\n2. from the start menu, launch *Command Prompt* and run the commands in [1]\n\n**WARNING:** Running these commands within an existing workspace directory will reset the dataset, deleting any \nexisting data. Ensure you have suitable backups of existing data.\n\n**Note:** Due to an internal limitation, on Windows computers the workspace directory must be on the same drive as\nthe Python interpreter. This will almost always be the `C:` drive.\n\n[1]\n\n**Note:** If viewing this documentation as a PDF, some commands below may disappear off the page. If you select the \nline you should be able to copy into a tool such as Notepad to view it in full.\n\n```\n$ C:\\ProgramData\\Miniconda3\\envs\\airnet\\Scripts\\activate.bat\n\n(airnet) $ C:\\ProgramData\\Miniconda3\\envs\\airnet\\python.exe C:\\ProgramData\\Miniconda3\\envs\\airnet\\Scripts\\airnet.exe init --dataset-path \'/path/to/workspace/directory\'\nDataset will be located at: \'/path/to/workspace/directory\'\n\nDataset created at: \'/path/to/workspace/directory\'\n```\n\n### Setup a Windows deployment VM\n\n**Note:** You will access to the BAS MAGIC resource pool in the BAS vSphere instance to complete these steps.\n\nTo create a test VM:\n\n1. login to the [BAS vSphere](https://bsv-vcsa-s1.nerc-bas.ac.uk/ui/) instance\n2. from the left panel, switch to the second view (to see templates)\n3. from the left panel, select *[root]* -> *Cambridge* -> *BASDEV* -> *air-unit-test-template*\n4. right-click and select *New VM from this template*:\n   1. select a suitable name (e.g. `air-unit-test5`) and choose the MAGIC folder/resource pool in the *BASDEV* cluster\n5. power on the resulting VM (the network may initially show as disconnected but will usually add itself)\n6. login as the IEUser with the shared password `Passw0rd!`.\n\nTo create the VM template used above:\n\n1. download the VMware version of the \n   [Microsoft Edge test VMs](https://developer.microsoft.com/en-us/microsoft-edge/tools/vms/)\n2. follow [these instructions](https://gitlab.data.bas.ac.uk/WSF/bas-base-images#vcentre) to create a VM template, \n   making changes as needed:\n   1. name the VM and template `air-unit-test`\n   2. set the RAM to 8GB\n   3. you should not need to add a virtual network interface or CD device, as this will already be setup\n\n## Development\n\n### Development environment\n\nGit and [Poetry](https://python-poetry.org) are required to set up a local development environment of this project.\n\n**Note:** If you use [Pyenv](https://github.com/pyenv/pyenv), this project sets a local Python version for consistency.\n\n```shell\n# clone from the BAS GitLab instance if possible\n$ git clone https://gitlab.data.bas.ac.uk/MAGIC/air-unit-network-dataset.git\n\n# setup virtual environment\n$ cd air-unit-network-dataset\n$ poetry install\n```\n\n### Running commands in development\n\nWithin a [Development Environment](#development-environment) use Poetry to test CLI commands. E.g.:\n\n```\n$ poetry run airnet --help\n```\n\n### Dependencies\n\nPython dependencies for this project are managed with [Poetry](https://python-poetry.org) in `pyproject.toml`.\n\n#### Adding new dependencies\n\nTo add a new (development) dependency:\n\n```shell\n$ poetry add [dependency] (--dev)\n```\n\n### Testing\n\nAutomated tests are not currently implemented (see \n[#16](https://gitlab.data.bas.ac.uk/MAGIC/air-unit-network-dataset/-/issues/16) for more information).\n\n#### Manual outputs testing\n\nManual testing of outputs is strongly encouraged prior to a release.\n\nThese should aim to be as comprehensive as possible, rather than only testing the outputs currently needed. This \ncurrently requires manual/ad-hoc modification of the `NetworkManager.dump_*()` methods as shown in this commit:\n\n* https://gitlab.data.bas.ac.uk/MAGIC/air-unit-network-dataset/-/blob/9892b56c1669bd88dfd5b7212c2d05861c7b5fa0/src/bas_air_unit_network_dataset/__init__.py#L1112\n* https://gitlab.data.bas.ac.uk/MAGIC/air-unit-network-dataset/-/blob/9892b56c1669bd88dfd5b7212c2d05861c7b5fa0/src/bas_air_unit_network_dataset/__init__.py#L1134\n\n## Deployment\n\nThe Air Unit Network utility is distributed in two forms:\n\n1. a Python package that can be installed through Pip from the\n   [PyPi](https://pypi.org/project/bas-air-unit-network-dataset/) registry\n2. a packaged Anaconda environment\n\n**Note:** Both distributions require OS dependencies to be installed separately (namely GDAL and LibXML2), see the \n[Installation](#installation) section for more information.\n\n### Python package\n\nThe Python package is standard, pure Python Pip package, built by Poetry as a binary wheel and source package.\n\nTo build the Python package manually:\n\n```\n$ poetry build\n```\n\nTo publish the Python package to PyPi manually, you will need an API token for the BAS organisational PyPi account,\nset as the `POETRY_PYPI_TOKEN_PYPI` environment variable. Then run:\n\n```\n$ poetry publish --username british-antarctic-survey\n```\n\n### Packaged Anaconda environment\n\nThe packaged Anaconda environment uses [Conda Pack](https://conda.github.io/conda-pack/) to create an OS/platform \nspecific archive of an Anaconda virtual environment, such that dependencies can be copied and installed offline. \nThis is designed for use in running the Air Unit Network utility in Antarctica.\n\n**Note:** This process will produce an environment that can be run on Windows (10), x86 64 bit computers only.\n\nFirst, [Create a Python Package](#python-package). Then within a \n[Windows Deployment VM](#setup-a-windows-deployment-vm):\n\n1. connect to OneDrive, such that the [Installation Bundle](#installation-bundle) can be accessed and updated\n2. from the Installation Bundle, run `miniconda-installer.exe`, installing for all users\n3. from the \'build\' directory in the Installation Bundle, run the 7Zip installer (`7z.exe`)\n4. from the start menu, launch *Anaconda PowerShell Prompt* with admin privileges and run the commands in [1]\n5. from the user *Downloads* directory, extract `airnet.tar.gz` using 7-zip: \n   1. then, extract `airnet.tar` to `airnet/`\n   2. then, right-click `airnet/` -> *Send to* -> *Compressed (zipped) folder* to produce `airnet.zip`\n   3. rename `airnet.zip` to `airnet-virtual-environment.zip`\n   4. copy `airnet.zip` to the definitive copy of the [Installation Bundle](#installation-bundle), replacing the \n      existing file\n\n[1]\n\n```shell\n(base) $ cd $env:userprofile\n(base) $ cd ./Downloads\n\n(base) $ conda create -n airnet\n(base) $ conda activate airnet\n(airnet) $ conda config --env --add channels conda-forge\n(airnet) $ conda config --env --set channel_priority strict\n(airnet) $ conda install fiona\n(airnet) $ python -m pip install bas-air-unit-network-dataset\n\n(airnet) $ conda activate base\n(base) $ conda install -c conda-forge conda-pack\n(base) $ conda pack -n airnet -o airnet.tar.gz\n```\n\n## Release procedure\n\nFor all releases:\n\n1. create a release branch\n2. [manually test outputs](#manual-outputs-testing)\n3. close release in `CHANGELOG.md`\n4. bump package version `poetry version [minor/patch]`\n5. build and publish [Python package](#python-package) to PyPi\n6. build a [Packaged Anaconda Environment](#packaged-anaconda-environment)\n7. convert project README to PDF using [markdowntopdf](https://www.markdowntopdf.com) and update in installation bundle\n8. push changes, merge the release branch into `main` and tag with version\n9. update the GitLab release with:\n    * a link to the relevant milestone(s)\n    * change log entries\n    * link to [PyPi package](https://pypi.org/project/bas-air-unit-network-dataset/#history)\n    * link to PDF version of documentation (share link to document in SharePoint)\n    * link to installation bundle (share link to archive in SharePoint)\n10. compress the [Installation Bundle](#installation-bundle) into a 7Zip archive (to allow fr transfer South via AMS) \n11. copy the [Installation Bundle](#installation-bundle) and [Test Network](#test-network) to a hard drive to take South\n\n## Feedback\n\nThe maintainer of this project is the BAS Mapping and Geographic Information Centre (MAGIC), they can be contacted at:\n[magic@bas.ac.uk](mailto:magic@bas.ac.uk).\n\n## Issue tracking\n\nThis project uses issue tracking, see the\n[Issue tracker](https://gitlab.data.bas.ac.uk/MAGIC/air-unit-network-dataset/-/issues) for more information.\n\n**Note:** Read & write access to this issue tracker is restricted. Contact the project maintainer to request access.\n\n## License\n\nCopyright (c) 2022 UK Research and Innovation (UKRI), British Antarctic Survey.\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the "Software"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE\nSOFTWARE.\n',
    'author': 'Felix Fennell',
    'author_email': 'felnne@bas.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/antarctica/bas-air-unit-network-dataset',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.1,<4.0.0',
}


setup(**setup_kwargs)
